use crate::components::game::cards::*;
use crate::components::websocket::WebSocketContext;
use leptos::prelude::*;
use lib::*;
use serde_json::to_string;
use thaw::*;

#[component]
pub fn JudgingView() -> impl IntoView {
    // Incoming
    let websocket = expect_context::<WebSocketContext>();
    let game_meta = expect_context::<ReadSignal<Option<GameMeta>>>();
    let judge_round = expect_context::<ReadSignal<Option<JudgeRound>>>();
    let set_judge_round = expect_context::<WriteSignal<Option<JudgeRound>>>();

    // Signals
    let (selected_cards, set_selected_cards) = signal::<Vec<WhiteCardMeta>>(vec![]);
    let (card_clicked, set_card_clicked) = signal::<String>(String::new());

    // Outoging
    provide_context::<WriteSignal<String>>(set_card_clicked);

    // On Incoming Judge
    Effect::new(move |_| {
        // Clear selected cards
        if judge_round().is_some() {
            set_selected_cards.update(|list| {
                list.clear();
            });
        }
    });

    // Card selection //
    // Toggle selected status of cards
    Effect::new(move |_| {
        if card_clicked() != "".to_string() {
            let identical_cards = selected_cards
                .get_untracked()
                .into_iter()
                .filter(|card| card.uuid == card_clicked.get_untracked())
                .collect::<Vec<WhiteCardMeta>>();

            if identical_cards.len() > 0 {
                set_selected_cards.update(|list| {
                    list.clear();
                });
            } else {
                // Clear selected cards
                set_selected_cards.update(|list| {
                    list.clear();
                });

                // Select card group
                for entry in judge_round.get_untracked().unwrap().entries {
                    for card in &entry.cards_to_judge {
                        if card.uuid == card_clicked() {
                            set_selected_cards.update(|cards| cards.extend(entry.cards_to_judge));
                            break;
                        }
                    }
                }
            }
            // Clear the signal otherwise it selects the last selected card again
            set_card_clicked.update_untracked(|value| value.clear());
        }
    });

    // Judging Submit Handler
    let submit_judge = move |_| {
        if !selected_cards().is_empty() {
            let msg = to_string(&JudgeDecisionRequest {
                game_id: game_meta().unwrap().uuid.clone(),
                winning_cards: selected_cards()
                    .iter()
                    .map(|card| card.uuid.clone())
                    .collect(),
            })
            .unwrap();

            websocket.send(&msg);
            set_selected_cards.update(|list| {
                list.clear();
            });

            set_judge_round(None);
        }
    };

    view! {
        <div class="w-full inline-flex flex-wrap">
            <BlackCard />
            // Selected cards
            {move || {
                selected_cards()
                    .into_iter()
                    .map(|card| {
                        view! { <WhiteCard card_data=card /> }
                    })
                    .collect_view()
            }}
        </div>

        // Submit button
        <div class="w-full inline-flex flex-wrap justify-center">
            <Button
                appearance=ButtonAppearance::Secondary
                button_type=ButtonType::Button
                on_click=submit_judge
            >
                Submit
            </Button>
        </div>

        <Show
            when=move || { judge_round().is_some() }
            fallback=move || {
                view! {
                    <p>
                        "You are the czar this round. Cards will appear here once all players have submitted their cards."
                    </p>
                }
            }
        >
            // Judge hand
            <div class="flex flex-nowrap overflow-scroll lg:inline-flex lg:flex-wrap lg:justify-center lg:overflow-none">
                {judge_round()
                    .unwrap()
                    .entries
                    .into_iter()
                    .map(|entries| {
                        view! {
                            <div class="inline-flex flex-wrap">
                                {entries
                                    .cards_to_judge
                                    .into_iter()
                                    .map(|card| {
                                        view! {
                                            <Show when={
                                                let waste_of_memory = card.clone();
                                                move || { !selected_cards().contains(&waste_of_memory) }
                                            }>
                                                <WhiteCard card_data=card.clone() />
                                            </Show>
                                        }
                                    })
                                    .collect_view()}
                            </div>
                        }
                    })
                    .collect_view()}
            </div>
        </Show>
    }
}
