use crate::game::*;
use anyhow::{Context, Result};
use lib::*;
use serde::Deserialize;
use std::{collections::HashMap, fs::read_to_string, sync::Arc};
use uuid::Uuid;

/// Card Set
pub struct CardSet {
    pub black: Option<Vec<Arc<CardBlackWithID>>>,
    pub white: Option<Vec<Arc<CardWhiteWithID>>>,
}

/// Card Packs
pub struct CardPacks {
    pub official: HashMap<u8, CardSet>,
    pub unofficial: HashMap<u8, CardSet>,
}

/// A raw white card as it exists in the source json
#[derive(Deserialize)]
struct CardWhiteFromJSON {
    /// Card text
    text: String,
    /// ID of the pack it came from
    pack: u8,
}

/// A raw black card as it exists in the source json
#[derive(Deserialize)]
struct CardBlackFromJSON {
    /// Card text
    text: String,
    /// Amount of cards to submit for judging
    pick: u8,
    /// ID of the pack it came from
    pack: u8,
}

/// A card pack
#[derive(Deserialize)]
struct CardPack {
    /// Name of the pack
    name: String,
    /// Whether or not this is an official card pack
    official: bool,
    /// White card data
    white: Option<Vec<CardWhiteFromJSON>>,
    /// Black card data
    black: Option<Vec<CardBlackFromJSON>>,
}

/// Parse json for card data
pub fn card_loader(
    path: &str,
) -> Result<(
    CardPacks,
    CardPacksMeta,
    // HashMap<String, Arc<CardWhiteWithID>>,
)> {
    // Load in json
    let data: String =
        read_to_string(path).with_context(|| format!("Invalid JSON path: \"{}\"", path))?;
    let jayson: Vec<CardPack> = serde_json::from_str(&data)
        .with_context(|| format!("The contents of \"{path}\" is not valid JSON."))?;

    // For global state
    let mut official: HashMap<u8, CardSet> = HashMap::new();
    let mut official_meta: Vec<CardPackMeta> = vec![];
    let mut unofficial: HashMap<u8, CardSet> = HashMap::new();
    let mut unofficial_meta: Vec<CardPackMeta> = vec![];
    // let mut white_cards_by_id = HashMap::<String, Arc<CardWhiteWithID>>::new();

    // Unpack the json
    for sets in jayson {
        let mut num_white = 0;
        let mut num_black = 0;

        let mut newset = CardSet {
            white: Option::None,
            black: Option::None,
        };

        // No safe default for this so make it an Option
        let mut pack: Option<u8> = Option::None;

        // Process white cards if there are any
        if let Some(ref white) = sets.white {
            num_white = white.len();
            if num_white > 0 {
                pack = Some(white[0].pack);
                let mut white_buf = vec![];
                for card in sets.white.unwrap() {
                    let uuid = Uuid::now_v7();

                    let new_card = Arc::new(CardWhiteWithID {
                        uuid,
                        text: card.text,
                    });

                    // white_cards_by_id.insert(uuid.to_string(), new_card.clone());
                    white_buf.push(new_card);
                }
                newset.white = Some(white_buf);
            }
        }

        // Process black cards if there are any
        if let Some(ref black) = sets.black {
            num_black = black.len();
            if num_black > 0 {
                pack = Some(black[0].pack);
                let mut black_buf = vec![];
                for card in sets.black.unwrap() {
                    black_buf.push(Arc::new(CardBlackWithID {
                        uuid: Uuid::now_v7(),
                        text: card.text,
                        pick: card.pick,
                    }));
                }
                newset.black = Some(black_buf);
            }
        }

        // Start repackaging
        let meta = CardPackMeta {
            name: sets.name,
            pack: pack.expect("No card pack number!").to_string(),
            num_white: num_white.try_into().unwrap(),
            num_black: num_black.try_into().unwrap(),
        };

        if sets.official {
            official_meta.push(meta);
            official.insert(pack.unwrap(), newset);
        } else {
            unofficial_meta.push(meta);
            unofficial.insert(pack.unwrap(), newset);
        }
    }

    // These are now the largest size they should ever be
    official.shrink_to_fit();
    unofficial.shrink_to_fit();
    official_meta.shrink_to_fit();
    unofficial_meta.shrink_to_fit();

    // Package for export
    let packs = CardPacks {
        official,
        unofficial,
    };
    let packs_meta = CardPacksMeta {
        official_meta,
        unofficial_meta,
    };

    Ok((packs, packs_meta))
}
