use crate::name_generator::*;
use crate::AppState;
use crate::DmUserMethod::*;
use crate::GameHandlerMessage;
use crate::OutgoingMessageHandlerMessage;
use crate::OutgoingMessageHandlerMessage::*;
use crate::SendUserMessage::*;
use crate::User;
use crate::UserHandlerMessage::*;
use axum::extract::ws::Message;
use lib::*;
use std::{
    collections::{HashMap, HashSet},
    net::SocketAddr,
    sync::{Arc, RwLock},
};
use tokio::sync::mpsc::Sender;

// TODO: clean up all this tx/msg mess

pub enum DmUserMethod {
    Addr(SocketAddr),
    Id(String),
}

/// For interacting with the user handler
pub enum UserHandlerMessage {
    NewUser(Sender<Message>, SocketAddr),
    UserLogIn(UserLogInRequest, SocketAddr),
    DmUser(SendUserMessage, DmUserMethod),
    Cleanup(SocketAddr),
    PassChatMessage(String, SocketAddr),
    GetPlayerUserId(JudgeDecisionRequest, SocketAddr),
    GetUser(NewGameRequest, SocketAddr),
    MoveRequest(PlayerMoveRequest, SocketAddr),
    JoinGame(String, SocketAddr),
}

/// Types of messages that can be sent to a user as a DM
// TODO: try to eliminate this extra step
pub enum SendUserMessage {
    SendUserUpdate(UserUpdate),
    SendChatMessage(ChatMessage),
    SendJudgeRound(JudgeRound),
}

/// Handles users
pub struct UserHandler {
    /// Pointer to global state
    state: Arc<AppState>,
    users_by_id: HashMap<String, Arc<RwLock<User>>>,
    name_generator: NameGenerator,
    reserved_names: HashSet<String>,
    offline_users: HashMap<String, Arc<RwLock<User>>>,
    online_users: HashMap<SocketAddr, Arc<RwLock<User>>>,
}

impl UserHandler {
    /// Returns new UserHandler
    pub fn new(state: Arc<AppState>) -> Self {
        let offline_users = HashMap::<String, Arc<RwLock<User>>>::new();
        let online_users = HashMap::<SocketAddr, Arc<RwLock<User>>>::new();
        UserHandler {
            offline_users,
            online_users,
            state,
            users_by_id: HashMap::<String, Arc<RwLock<User>>>::new(),
            reserved_names: HashSet::<String>::new(),
            name_generator: NameGenerator::new(),
        }
    }

    /// Handles incoming messages
    pub async fn handle(&mut self, message: UserHandlerMessage) {
        match message {
            NewUser(dm_tx, addr) => self.set_up_new_user(dm_tx, addr).await,
            UserLogIn(request, addr) => self.login(request, addr).await,
            DmUser(message, method) => self.dm_user(message, method).await,
            Cleanup(addr) => self.user_cleanup(addr).await,
            PassChatMessage(msg, addr) => self.pass_chat_message(msg, addr).await,
            GetPlayerUserId(judge_decision, addr) => {
                self.get_player_user_id(judge_decision, addr).await
            }
            GetUser(new_game, addr) => self.get_user(new_game, addr).await,
            MoveRequest(move_request, addr) => self.move_request(move_request, addr).await,
            JoinGame(game_id, addr) => self.join_game(game_id, addr).await,
        }
    }

    async fn join_game(&self, game_id: String, addr: SocketAddr) {
        let user = self.online_users.get(&addr).unwrap().clone();

        if let Err(e) = self
            .state
            .tx_game_handler
            .send(GameHandlerMessage::JoinGame(game_id, user))
            .await
        {
            tracing::error!("Error: {}", e)
        }
    }

    async fn move_request(&self, move_request: PlayerMoveRequest, addr: SocketAddr) {
        let player_user = self.online_users.get(&addr).unwrap().clone();
        if let Err(e) = self
            .state
            .tx_game_handler
            .send(GameHandlerMessage::MoveRequest(move_request, player_user))
            .await
        {
            tracing::error!("Error: {}", e)
        }
    }

    async fn get_user(&self, new_game: NewGameRequest, addr: SocketAddr) {
        let host = self.online_users.get(&addr).unwrap().clone();
        if let Err(e) = self
            .state
            .tx_game_handler
            .send(GameHandlerMessage::NewGame(new_game, host))
            .await
        {
            tracing::error!("Error requesting new game: {}", e)
        }
    }

    async fn get_player_user_id(&self, judge_decision: JudgeDecisionRequest, addr: SocketAddr) {
        let player_user_id = self
            .online_users
            .get(&addr)
            .unwrap()
            .read()
            .unwrap()
            .uuid
            .clone();

        if let Err(e) = self
            .state
            .tx_game_handler
            .send(GameHandlerMessage::JudgeDecision(
                judge_decision,
                player_user_id,
            ))
            .await
        {
            tracing::error!("Error contacting game handler: {}", e)
        }
    }

    async fn pass_chat_message(&self, msg: String, addr: SocketAddr) {
        let text =
            format! {"{0}: {1}", self.online_users.get(&addr).unwrap().read().unwrap().name, msg};
        if let Err(e) = self
            .state
            .tx_outgoing_message_handler
            .send(OutgoingMessageHandlerMessage::Broadcast(
                ServerToClientMessage::ChatMessage(ChatMessage { text }),
            ))
            .await
        {
            tracing::error!("Can't contact outgoing message handler: {}", e);
        }
    }

    /// Send message direct to a single user
    async fn dm_user(&self, message: SendUserMessage, method: DmUserMethod) {
        let tx;

        // Determine lookup method
        match method {
            Id(id) => {
                if let Some(user) = self.users_by_id.get(&id) {
                    tx = user.read().unwrap().tx.clone();
                } else {
                    tracing::error!("Attempted to send message to invalid user id!");
                    // TODO: User can still be offline causing send to fail. This is suppressed
                    // below
                    return;
                }
            }
            Addr(addr) => {
                if let Some(user) = self.online_users.get(&addr) {
                    tx = user.read().unwrap().tx.clone();
                } else {
                    tracing::error!("Attempted to send message to offline user!");
                    return;
                }
            }
        }

        // Serialize and send message
        match message {
            SendUserUpdate(message) => {
                if let Err(e) = self
                    .state
                    .tx_outgoing_message_handler
                    .send(Unicast((tx, ServerToClientMessage::UserUpdate(message))))
                    .await
                {
                    tracing::error!("Error sending user update: {}", e)
                }
            }
            SendChatMessage(message) => {
                if let Err(e) = self
                    .state
                    .tx_outgoing_message_handler
                    .send(Unicast((tx, ServerToClientMessage::ChatMessage(message))))
                    .await
                {
                    tracing::error!("Error sending chat message: {}", e)
                }
            }
            SendJudgeRound(message) => {
                if let Err(e) = self
                    .state
                    .tx_outgoing_message_handler
                    .send(Unicast((tx, ServerToClientMessage::JudgeRound(message))))
                    .await
                {
                    tracing::error!("Error sending judge round: {}", e)
                }
            }
        }
    }

    /// Create, register, and hydrate new user
    async fn set_up_new_user(&mut self, dm_tx: Sender<Message>, addr: SocketAddr) {
        let name = self.name_generator.generate(&self.reserved_names);

        let new_user = Arc::new(RwLock::new(User::new(name, dm_tx.clone())));

        // Register uuid
        self.users_by_id
            .insert(new_user.read().unwrap().uuid.clone(), new_user.clone());

        // Register online using `addr` as key until something longer lived exists
        self.set_user_online(addr, new_user.clone());

        // Hydrate client
        // Notify client of username
        if let Err(e) = self
            .state
            .tx_outgoing_message_handler
            .send(Unicast((
                dm_tx.clone(),
                ServerToClientMessage::UserUpdate(user_client_self_update(&new_user)),
            )))
            .await
        {
            tracing::error!("Error sending client update: {}", e)
        }

        // Send game update
        // TODO: send single update instad of broadcasting to all users
        if let Err(e) = self
            .state
            .tx_game_handler
            .send(GameHandlerMessage::BroadcastGamesUpdate())
            .await
        {
            tracing::error!("Error contacing game handler {}", e)
        }

        // Send card packs for new game form
        if let Err(e) = self
            .state
            .tx_game_handler
            .send(GameHandlerMessage::SendCardPacks(dm_tx.clone()))
            .await
        {
            tracing::error!("Error contacing game handler {}", e)
        }

        // Broadcast new user's existence
        // TODO: this should probably be combined and sent as one
        if let Err(e) = self
            .state
            .tx_outgoing_message_handler
            .send(Broadcast(ServerToClientMessage::ChatMessage(
                self.meta_announce_user_join(&addr),
            )))
            .await
        {
            tracing::error!("Error contacting outgoing message handler: {}", e);
        }

        self.broadcast_user_count();

        self.broadcast_chat_update().await;

        // Send motd
        // TODO: this races the broadcasts but if it's done last it'll probably show up last...

        if let Err(e) = self
            .state
            .tx_outgoing_message_handler
            .send(Unicast((
                dm_tx.clone(),
                ServerToClientMessage::ChatMessage(meta_motd()),
            )))
            .await
        {
            tracing::error!("Error contacting outgoing message handler: {}", e);
        }
    }

    /// Handle user login
    async fn login(&mut self, request: UserLogInRequest, addr: SocketAddr) {
        let username_max_len = 66; // This is the longest name the generator may produce right now
        let new_name;

        if request.username.len() > username_max_len {
            new_name = request.username[..username_max_len].to_string();
        } else {
            new_name = request.username;
        }

        let old_name;

        if let Some(user) = self.online_users.get(&addr) {
            old_name = user.read().unwrap().name.clone();

            // User abandons current name by requesting a new one
            self.reserved_names.remove(&old_name);
        } else {
            tracing::error!("Nonexistent user attempting login");
            return;
        }

        // Resume user's old session if they exist as offline
        if self.offline_users.contains_key(&new_name) {
            let buf;
            // Copy over new tx
            if let Some(online_user) = self.online_users.remove(&addr) {
                if let Some(offline_user) = self.offline_users.remove(&new_name) {
                    offline_user.write().unwrap().tx = online_user.write().unwrap().tx.clone();
                    buf = offline_user;
                } else {
                    tracing::error!("Error copying tx to new user: Can't find offline user!");
                    return;
                }
            } else {
                tracing::error!("Error copying tx to new user: Can't find online user!");
                return;
            }

            // Move offline user object to online
            self.online_users.insert(addr, buf);

            // Send welcome back messages
            let msg = format! {
                "{0} changed name to {1}. Welcome back!",
                old_name,
                new_name
            };

            if let Err(e) = self
                .state
                .tx_outgoing_message_handler
                .send(Broadcast(ServerToClientMessage::ChatMessage(ChatMessage {
                    text: msg,
                })))
                .await
            {
                tracing::error!("Error contacting outgoing message handler: {}", e);
            }
        }
        // Check if name is taken by an online user
        else if self.reserved_names.contains(&new_name) {
            self.dm_user(
                SendChatMessage(ChatMessage {
                    text: "Name is taken".to_string(),
                }),
                Addr(addr),
            )
            .await;
            self.dm_user(
                SendUserUpdate(UserUpdate {
                    username: old_name.clone(),
                }),
                Addr(addr),
            )
            .await;
            return;
        } else {
            // Reserve name
            self.reserved_names.insert(new_name.clone());

            // Change user's name
            if let Some(user) = self.online_users.get_mut(&addr) {
                user.write().unwrap().change_name(new_name.clone());
            } else {
                tracing::error!("Error updating username: Can't find user!");
                return;
            }

            // Send chat updates
            let msg = format! {
                "{0} changed name to {1}.",
                old_name,
                new_name
            };

            if let Err(e) = self
                .state
                .tx_outgoing_message_handler
                .send(Broadcast(ServerToClientMessage::ChatMessage(ChatMessage {
                    text: msg,
                })))
                .await
            {
                tracing::error!("Error contacting outgoing message handler: {}", e);
            }
        }

        // Send the user their new name
        self.dm_user(
            SendUserUpdate(UserUpdate {
                username: new_name.clone(),
            }),
            Addr(addr),
        )
        .await;
        // Update games this user is in
        if let Some(user) = &self.online_users.get(&addr) {
            let user_id = user.read().unwrap().uuid.to_string();
            let msg = GameHandlerMessage::SendGameUserUpdate(user_id);
            let tx = self.state.tx_game_handler.clone();
            tokio::spawn(async move { tx.send(msg).await });
        }
        // Send client updates
        let tx = self.state.tx_game_handler.clone();
        let msg = GameHandlerMessage::BroadcastGamesUpdate();
        tokio::spawn(async move { tx.send(msg).await });

        self.broadcast_chat_update().await;
    }

    /// Broadcast updated user count
    fn broadcast_user_count(&self) {
        let tx = self.state.tx_outgoing_message_handler.clone();
        let online_users: u32 = self.online_users.len().try_into().unwrap();
        let msg = ServerToClientMessage::ServerOnlineUsers(ServerOnlineUsers { online_users });
        tokio::spawn(async move { tx.send(Broadcast(msg)).await });
    }

    /// Clean up after a user when they disconnect
    async fn user_cleanup(&mut self, addr: SocketAddr) {
        let user_name = self
            .online_users
            .get(&addr)
            .unwrap()
            .read()
            .unwrap()
            .name
            .clone();

        // Announce User left in chat
        let msg = ChatMessage {
            text: format!("{0} left.", &user_name),
        };

        if let Err(e) = self
            .state
            .tx_outgoing_message_handler
            .send(Broadcast(ServerToClientMessage::ChatMessage(msg)))
            .await
        {
            tracing::error!("Error contacting outgoing message handler: {}", e);
        }

        // Move user to offline
        self.set_user_offline(user_name, &addr);

        self.broadcast_user_count();

        self.broadcast_chat_update().await;
    }

    /// Set user status to online
    fn set_user_online(&mut self, addr: SocketAddr, user: Arc<RwLock<User>>) {
        self.online_users.insert(addr, user);
    }

    /// Set user status to offline
    fn set_user_offline(&mut self, user_name: String, addr: &SocketAddr) {
        self.offline_users
            .insert(user_name, self.online_users.remove(addr).unwrap());
    }

    /// Update chat stuff like users list
    async fn broadcast_chat_update(&self) {
        if let Err(e) = self
            .state
            .tx_outgoing_message_handler
            .send(Broadcast(ServerToClientMessage::ChatUpdate(
                self.generate_chat_update(),
            )))
            .await
        {
            tracing::error!("Error contacting outgoing message handler: {}", e);
        }
    }
    /// Generate chatroom join announcement
    fn meta_announce_user_join(&self, addr: &SocketAddr) -> ChatMessage {
        let msg = format!("{} joined.", {
            if let Some(user) = self.online_users.get(addr) {
                user.read().unwrap().name.clone()
            } else {
                "Error!".to_string()
            }
        });

        ChatMessage { text: msg }
    }
    /// Generate chatroom metadata update
    fn generate_chat_update(&self) -> ChatUpdate {
        // TODO: this may get expensive if there are many users
        let mut names = vec![];

        for user in self.online_users.iter() {
            names.push(user.1.read().unwrap().name.clone());
        }

        ChatUpdate {
            room: "Lobby".to_string(),
            users: names,
        }
    }
}

/// Generate message to notify client of user changes
pub fn user_client_self_update(new_user: &Arc<RwLock<User>>) -> UserUpdate {
    UserUpdate {
        username: new_user.read().unwrap().name.clone(),
    }
}

/// Generate message-of-the-day server greeting
pub fn meta_motd() -> ChatMessage {
    ChatMessage {
        text: "Greetings from the game server!".to_string(),
    }
}
